import numpy as np
import torch
import torch.nn as nn

from ..core import (
    ForecastingModule,
    Head,
)

from ..chunk import (
    EncodingChunkSpec,
    DecodingChunkSpec,
    LabelChunkSpec,
)


class RNN(ForecastingModule):
    def __init__(
        self,
        hidden_size,
        encoding_length,
        decoding_length,
        target_names,
        nontarget_names,
        n_layers,
        rnn_class,
        dropout_rate=0.0,
        lr=1e-3,
        optimizer: torch.optim.Optimizer = torch.optim.Adam,
        optimizer_options=None,
        loss_fn=None,
        metrics=None,
        head=None,
    ):
        super().__init__()
        self.save_hyperparameters()

        self.encoding_length = encoding_length
        self.decoding_length = decoding_length

        if optimizer_options is None:
            self.hparams.optimizer_options = {}

        if loss_fn is None:
            loss_fn = nn.MSELoss()

        n_outputs = len(target_names)
        n_features = len(nontarget_names) + n_outputs

        self.use_nontargets = n_outputs != n_features

        self.encoder = rnn_class(
            input_size=n_features,
            hidden_size=hidden_size,
            num_layers=n_layers,
            batch_first=True,
            dropout=dropout_rate,
        )

        # Use same model instance used in encoding process.
        self.decoder = self.encoder

        if head is not None:
            self.head = head
        else:
            self.head = Head(
                tag='targets',
                output_module=nn.Linear(hidden_size, n_outputs),
                loss_fn=loss_fn,
                metrics=metrics,
            )

    def encode(self, inputs):
        # (B, L, F).
        if self.use_nontargets:
            x = torch.cat([
                inputs['encoding.targets'],
                inputs['encoding.nontargets']
            ], dim=2)
        else:
            x = inputs['encoding.targets']

        # (B, L, H).
        # For LSTM, hidden_state is a tuple: (h_0, c_0).
        h, memory = self.encoder(x)

        return {
            'h': h[:, -1:, :],
            'memory': memory,
        }

    def decode(self, inputs):
        self.head.reset()
        y = self.head(inputs['h'])

        if self.use_nontargets:
            c = inputs['decoding.nontargets']
            x = torch.cat([y, c[:, 0:1, :]], dim=2)
        else:
            x = y

        memory = inputs['memory']

        for i in range(1, self.decoding_length):
            h, memory = self.decoder(x, memory)
            y = self.head(h)

            if i+1 == self.decoding_length:
                break

            if self.use_nontargets:
                x = torch.cat([y, c[:, i:i+1, :]], dim=2)
            else:
                x = y

        outputs = self.head.get_outputs()

        return outputs

    def make_chunk_specs(self):
        E = self.encoding_length
        D = self.decoding_length

        chunk_specs = [
            EncodingChunkSpec(
                tag='targets',
                names=self.hparams.target_names,
                range_=(0, E),
                dtype=np.float32
            ),
            LabelChunkSpec(
                tag='targets',
                names=self.hparams.target_names,
                range_=(E, E+D),
                dtype=np.float32,
            ),
        ]

        if self.use_nontargets:
            chunk_specs += [
                EncodingChunkSpec(
                    tag='nontargets',
                    names=self.hparams.nontarget_names,
                    range_=(1, E+1),
                    dtype=np.float32,
                ),
                DecodingChunkSpec(
                    tag='nontargets',
                    names=self.hparams.nontarget_names,
                    range_=(E+1, E+D),
                    dtype=np.float32,
                ),
            ]

        return chunk_specs

    def configure_optimizers(self):
        return self.hparams.optimizer(
            self.parameters(),
            lr=self.hparams.lr,
            **self.hparams.optimizer_options,
        )


import numpy as np
import xarray as xr

import torch


class BaseChunkSpec:
    PREFIX = ''

    def __init__(self, tag, names, range_, dtype):
        self.tag = tag
        self.names = names
        self.range = range_
        self.dtype = dtype

    @property
    def tag(self) -> str:
        if self._tag is None:
            raise NotImplementedError(
                f'Define {self.__class__._name__}.tag'
            )

        return self._tag

    @tag.setter
    def tag(self, value: str):
        if not isinstance(value, str):
            raise TypeError(
                f'Invalid type for "tag": {type(value)}'
            )

        if not value.startswith(self.PREFIX):
            value = f'{self.PREFIX}.{value}'

        self._tag = value

    @property
    def names(self) -> list[str]:
        if self._names is None:
            raise NotImplementedError(
                f'Define {self.__class__._name__}.names'
            )

        return self._names

    @names.setter
    def names(self, value):
        if not isinstance(value, (list, tuple)):
            raise TypeError(
                f'Invalid type for "names"'
            )
        if not all((isinstance(name, str) for name in value)):
            raise TypeError(
                f'Invalid type for "names"'
            )

        self._names = list(value)

    @property
    def range(self) -> tuple[int, int]:
        if self._range is None:
            raise NotImplementedError(
                f'Define {self.__class__._name__}.range'
            )

        return self._range

    @range.setter
    def range(self, value: tuple[int, int]):
        if not isinstance(value, (tuple, list)):
            raise TypeError(
                f'Invalid type for "range": {type(value)}'
            )

        if not isinstance(value[0], int):
            raise TypeError(
                f'Invalid type for "range[0]": {type(value)}'
            )

        if not isinstance(value[1], int):
            raise TypeError(
                f'Invalid type for "range[1]": {type(value)}'
            )

        if value[0] >= value[1]:
            raise ValueError('range[0] >= range[1]')

        self._range = tuple(value)

    @property
    def dtype(self) -> np.dtype:
        if self._dtype is None:
            raise NotImplementedError(
                f'Define {self.__class__._name__}.dtype'
            )

        return self._dtype

    @dtype.setter
    def dtype(self, value: np.dtype):
        if not isinstance(np.dtype(value), np.dtype):
            raise TypeError(
                f'Invalid type for "dtype": {type(value)}'
            )

        self._dtype = np.dtype(value)


class EncodingChunkSpec(BaseChunkSpec):
    PREFIX = 'encoding'


class DecodingChunkSpec(BaseChunkSpec):
    PREFIX = 'decoding'


class LabelChunkSpec(BaseChunkSpec):
    PREFIX = 'label'


class ChunkExtractor:
    def __init__(self, df, chunk_specs):
        # Check tag duplication.
        tags = [spec.tag for spec in chunk_specs]
        if len(tags) != len(set(tags)):
            raise ValueError(
                f'Tags are duplicated. {[s.tag for s in chunk_specs]}.'
            )

        self.chunk_specs = chunk_specs
        self.chunk_min_t = min(spec.range[0] for spec in chunk_specs)
        self.chunk_max_t = max(spec.range[1] for spec in chunk_specs)
        self.chunk_length = self.chunk_max_t - self.chunk_min_t

        self.time_index_values = np.arange(len(df))

        self._preprocess(df)

    def _preprocess(self, df):
        self.data = {}
        for spec in self.chunk_specs:
            values = df[spec.names].astype(spec.dtype).values
            if len(values.shape) == 1:
                values = values[:, np.newaxis]
            self.data[spec.tag] = values

    def extract(self, start_time_index, return_time_index=False):
        assert start_time_index + self.chunk_min_t >= 0

        times = self.time_index_values[
            start_time_index + self.chunk_min_t :
            start_time_index + self.chunk_max_t
        ]

        chunk_dict = {}
        for spec in self.chunk_specs:
            array = self.data[spec.tag][
                start_time_index + self.chunk_min_t :
                start_time_index + self.chunk_max_t
            ]

            range_ = (
                spec.range[0] - self.chunk_min_t,
                spec.range[1] - self.chunk_min_t,
            )

            chunk_dict[spec.tag] = array[slice(*range_)]
            # Time index information.
            if return_time_index:
                chunk_dict[f'{spec.tag}.time_index'] = times[slice(*range_)]

        return chunk_dict


class ChunkInverter:
    def __init__(self, chunk_specs: list[BaseChunkSpec]):
        """Class to convert tensor to DataFrame."""
        self.chunk_specs = chunk_specs

        # Core tag to names dict.
        self.core_tag_dict = {}
        for spec in chunk_specs:
            core_tag = spec.tag.split('.')[1]

            if core_tag not in self.core_tag_dict:
                self.core_tag_dict[core_tag] = spec.names
            else:
                a = np.array(self.core_tag_dict[core_tag])
                b = np.array(spec.names)
                assert np.all(a ==  b)

        self.tag_dict = {spec.tag: spec.names for spec in chunk_specs}

    def _convert_to_numpy(self, tensor: torch.Tensor | np.ndarray):
        if isinstance(tensor, np.ndarray):
            return tensor
        elif isinstance(tensor, torch.Tensor):
            return tensor.cpu().numpy()
        else:
            raise TypeError(f'Invalid type for tensor {type(tensor)}')

    def _infer_shape(self, tensor: np.ndarray):
        pass

    def invert(self, tag: str, tensor: torch.Tensor | np.ndarray):
        if tag in self.tag_dict:
            names = self.tag_dict[tag]
        elif tag in self.core_tag_dict:
            names = self.core_tag_dict[tag]
        elif tag.split('.')[1] in self.core_tag_dict:
            names = self.core_tag_dict[tag.split('.')[1]]
        else:
            raise ValueError(f'"{tag}" not in chunk_specs.')

        data = self._convert_to_numpy(tensor)

        da = xr.DataArray(
            data=data,
            dims=('batch_index', 'time_index', 'feature'),
            name='',
        )

        # TODO: support for multi-dimensional features (like video).

        df = da.to_dataframe().unstack(2)
        df.columns = names

        return df

    def invert_dict(
        self,
        data: dict[str, torch.Tensor | np.ndarray],
    ):
        outputs = {}
        for tag, tensor in data.items():
            print(tag)
            outputs[tag] = self.invert(tag, tensor)

        return outputs


import numpy as np
import torch
import torch.nn as nn

from ..core import (
    ForecastingModule,
    Head,
)

from ..chunk import (
    EncodingChunkSpec,
    DecodingChunkSpec,
    LabelChunkSpec,
)

from ..layer import PositionalEncoding


class SingleShotTransformer(ForecastingModule):
    def __init__(
        self,
        encoding_length,
        decoding_length,
        target_names,
        nontarget_names,
        d_model,
        n_heads,
        n_layers,
        dim_feedforward=None,
        dropout_rate=0.0,
        lr=1e-3,
        optimizer: torch.optim.Optimizer = torch.optim.Adam,
        optimizer_options=None,
        loss_fn=None,
        metrics=None,
        head=None,
    ):
        super().__init__()
        self.save_hyperparameters()

        self.encoding_length = encoding_length
        self.decoding_length = decoding_length

        if optimizer_options is None:
            self.hparams.optimizer_options = {}

        if loss_fn is None:
            loss_fn = nn.MSELoss()

        if dim_feedforward is None:
            dim_feedforward = 4 * d_model

        n_targets = len(target_names)
        n_nontargets = len(nontarget_names)
        n_features = n_nontargets + n_targets

        self.use_nontargets = n_nontargets > 0

        self.encoder_d_matching_layer = nn.Linear(
            in_features=n_features,
            out_features=d_model,
        )

        if self.use_nontargets:
            self.decoder_d_matching_layer = nn.Linear(
                in_features=n_nontargets,
                out_features=d_model,
            )

        self.positional_encoding = PositionalEncoding(
            d_model=d_model, max_len=max(encoding_length, decoding_length),
        )

        encoder_layer = nn.TransformerEncoderLayer(
            d_model, n_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout_rate,
            batch_first=True,
        )

        self.encoder = nn.TransformerEncoder(encoder_layer, n_layers)

        decoder_layer = nn.TransformerDecoderLayer(
            d_model, n_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout_rate,
            batch_first=True,
        )

        self.decoder = nn.TransformerDecoder(decoder_layer, n_layers)

        if head is not None:
            self.head = head
        else:
            self.head = Head(
                tag='targets',
                output_module=nn.Linear(d_model, n_targets),
                loss_fn=loss_fn,
                metrics=metrics,
            )

    def encode(self, inputs):
        # L: encoding length.
        # all_input: (B, L, F).
        if self.use_nontargets:
            x = torch.cat([
                inputs['encoding.targets'],
                inputs['encoding.nontargets']
            ], dim=2)
        else:
            x = inputs['encoding.targets']


        x = self.encoder_d_matching_layer(x)
        x = self.positional_encoding(x)

        # (B, L, d_model).
        memory = self.encoder(x)

        return {
            'memory': memory
        }

    def decode(self, inputs):
        # L: decoding_length
        memory = inputs['memory']

        if self.use_nontargets:
            x = inputs['decoding.nontargets']
            x = self.decoder_d_matching_layer(x)
        else:
            # Same device will be used automatically.
            x = torch.zeros_like(memory)

        x = self.positional_encoding(x)

        # (B, L, d_model).
        tgt_mask = self.generate_square_subsequent_mask(x.size(1))
        x = self.decoder(tgt=x, memory=memory, tgt_mask=tgt_mask)

        # (B, L, n_outputs).
        self.head.reset()
        self.head(x)
        outputs = self.head.get_outputs()

        return outputs

    def generate_square_subsequent_mask(self, sz):
        r"""Generate a square mask for the sequence.
            The masked positions are filled with float('-inf').
            Unmasked positions are filled with float(0.0).
        """
        return torch.triu(
            torch.full((sz, sz), float('-inf')), diagonal=1
        ).to(self.device)

    def make_chunk_specs(self):
        E = self.encoding_length
        D = self.decoding_length

        chunk_specs = [
            EncodingChunkSpec(
                tag='targets',
                names=self.hparams.target_names,
                range_=(0, E),
                dtype=np.float32
            ),
            LabelChunkSpec(
                tag='targets',
                names=self.hparams.target_names,
                range_=(E, E+D),
                dtype=np.float32,
            ),
        ]

        if self.use_nontargets:
            chunk_specs += [
                EncodingChunkSpec(
                    tag='nontargets',
                    names=self.hparams.nontarget_names,
                    range_=(0, E),
                    dtype=np.float32,
                ),
                DecodingChunkSpec(
                    tag='nontargets',
                    names=self.hparams.nontarget_names,
                    range_=(E, E+D),
                    dtype=np.float32,
                ),
            ]

        return chunk_specs

    def configure_optimizers(self):
        return self.hparams.optimizer(
            self.parameters(),
            lr=self.hparams.lr,
            **self.hparams.optimizer_options,
        )


import copy
import itertools

import pandas as pd


def logical_and_for_set_list(set_list):
    s = set_list[0]
    for s_ in set_list[1:]:
        s = s & s_
    return s


def logical_or_for_set_list(set_list):
    s = set_list[0]
    for s_ in set_list[1:]:
        s = s | s_
    return s


def merge_dicts(dicts, ignore_keys=None):
    # Check key duplication.
    assert len(
        logical_and_for_set_list([set(d.keys()) for d in dicts])
    ) == 0, 'Keys of dictionaries are duplicated.'

    if ignore_keys is None:
        ignore_keys = set()
    else:
        ignore_keys = set(ignore_keys)

    merged_dict = {}
    for k, v in itertools.chain(*[d.items() for d in dicts]):
        if k in ignore_keys:
            continue
        merged_dict[k] = v

    return merged_dict


def merge_data_frames(dfs):
    dfs = copy.deepcopy(dfs)
    for i, df in enumerate(dfs):
        df['time_index'] = df.index
        df['time_series_id'] = i

    return pd.concat(dfs).reset_index(drop=True)


import matplotlib as mpl
import matplotlib.pyplot as plt


def plot_chunks(chunk_specs):
    # ys = [0, 0]
    # widths = [encoding_length, decoding_length]
    # lefts = [0, encoding_length]
    # tags = ['encoding', 'decoding']
    ys = []
    widths = []
    lefts = []
    tags = []

    for y, spec in enumerate(chunk_specs, start=1):
        ys.append(y)
        widths.append(spec.range[1]-spec.range[0])
        lefts.append(spec.range[0])
        tags.append(spec.tag)

    for y, width, left, tag in zip(ys, widths, lefts, tags):
        plt.barh(y=y, width=width, left=left, alpha=0.8)
        plt.annotate(tag, (left, y))


import torch
import torch.nn as nn
import torch.distributions

import pytorch_lightning as pl

from collections import defaultdict
from typing import Any, Callable
from torchmetrics import Metric, MetricCollection
from .util import merge_dicts


class MetricModule(nn.Module):
    def __init__(
        self,
        tag: str,
        metrics: Metric | list[Metric] | dict[str, Metric]
    ):
        super().__init__()

        self.tag = tag

        if tag.startswith('head.'):
            self.head_tag = tag
            self.label_tag = f'label.{tag[5:]}'
        else:
            self.head_tag = f'head.{tag}'
            self.label_tag = f'label.{tag}'

        metrics = MetricCollection(metrics)

        self._metric_dict = nn.ModuleDict({
            '_train': metrics.clone(prefix=f'train/{tag}.'),
            '_val': metrics.clone(prefix=f'val/{tag}.'),
            '_test': metrics.clone(prefix=f'test/{tag}.')
        })

    def forward(self, outputs, batch, stage):
        return self._metric_dict[f'_{stage}'](
            outputs[self.head_tag], batch[self.label_tag]
        )

    def compute(self, stage):
        return self._metric_dict[f'_{stage}'].compute()

    def update(self, outputs, batch, stage):
        self._metric_dict[f'_{stage}'].update(
            outputs[self.head_tag], batch[self.label_tag]
        )

    def reset(self, stage):
        self._metric_dict[f'_{stage}'].reset()


class BaseHead(nn.Module):
    _SPECIAL_ATTRIBUTES = (
        'tag',
        'loss_weight',
        'metrics'
    )

    def __init__(self):
        """Base class of all Head classes."""
        super().__init__()
        self._tag = None
        self._loss_weight = None
        self._metrics = None

    def __setattr__(self, name, value):
        if name in BaseHead._SPECIAL_ATTRIBUTES:
            return object.__setattr__(self, name, value)
        else:
            return super().__setattr__(name, value)

    @property
    def tag(self) -> str:
        """Tag for a head. Prefix 'head.' is added automatically."""
        if self._tag is None:
            raise NotImplementedError(
                f'Define {self.__class__.__name__}.tag'
            )
        else:
            return self._tag

    @tag.setter
    def tag(self, value: str):
        if not isinstance(value, str):
            raise TypeError(
                f'Invalid type for "tag": {type(value)}'
            )

        if not value.startswith('head.'):
            value = f'head.{value}'

        self._tag = value

    @property
    def metrics(self) -> MetricModule:
        if self._metrics is None:
            raise NotImplementedError(
                f'Define {self.__class__.__name__}.metrics'
            )
        else:
            return self._metrics

    @metrics.setter
    def metrics(self, value: Metric | list[Metric] | dict[str, Metric]):
        metric_module = MetricModule(tag=self.tag, metrics=value)
        self._metrics = metric_module

    @property
    def has_metrics(self):
        return self._metrics is not None

    @property
    def loss_weight(self) -> float:
        """Loss weight for loss calculations."""
        if self._loss_weight is None:
            self._loss_weight = 1.0
        return self._loss_weight

    @loss_weight.setter
    def loss_weight(self, value: float):
        if not isinstance(value, (float, int)):
            raise TypeError(
                f'Invalid type for "loss_weight": {type(value)}'
            )
        elif value < 0:
            raise ValueError('loss_weight < 0')

        self._loss_weight = value

    @property
    def label_tag(self) -> str:
        """Tag of target label. If the tag of head is "head.my_tag" then
        label_tag is "label.my_tag".
        """
        return f'label.{self.tag[5:]}'

    def forward(self, inputs: Any) -> torch.Tensor:
        raise NotImplementedError(
            f'Define {self.__class__.__name__}.forward()'
        )

    def get_outputs(self) -> dict[str, Any]:
        raise NotImplementedError(
            f'Define {self.__class__.__name__}.get_outputs()'
        )

    def reset(self):
        raise NotImplementedError(
            f'Define {self.__class__.__name__}.reset()'
        )

    def calculate_loss(
        self,
        outputs: dict[str, Any],
        batch: dict[str, Any]
    ) -> torch.Tensor:
        raise NotImplementedError(
            f'Define {self.__class__.__name__}.calculate_loss()'
        )


class Head(BaseHead):
    def __init__(
        self,
        tag: str,
        output_module: nn.Module,
        loss_fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        loss_weight: float = 1.0,
        metrics: Metric | list[Metric] | dict[str, Metric] = None,
    ):
        super().__init__()

        self.tag = tag
        self.output_module = output_module
        self.loss_fn = loss_fn
        self.loss_weight = loss_weight

        if metrics is not None:
            self.metrics = metrics

        self._ys = []

    def forward(self, inputs: Any) -> torch.Tensor:
        y = self.output_module(inputs)
        self._ys.append(y)
        return y

    def get_outputs(self):
        return {
            self.tag: torch.cat(self._ys, dim=1)
        }

    def reset(self):
        self._ys = []

    def calculate_loss(
        self,
        outputs: dict[str, Any],
        batch: dict[str, Any]
    ) -> torch.Tensor:
        return self.loss_fn(outputs[self.tag], batch[self.label_tag])


class DistributionHead(BaseHead):
    def __init__(self,
        tag: str,
        distribution: torch.distributions.Distribution,
        in_features: int,
        out_features: int,
        loss_weight: float = 1.0,
        metrics: Metric | list[Metric] | dict[str, Metric] = None,
    ):
        super().__init__()

        self.tag = tag
        self.loss_weight = loss_weight
        if metrics is not None:
            self.metrics = metrics

        linears = {}
        transforms = {}
        for k, v in distribution.arg_constraints.items():
            # 'logits' is prefered.
            if k == 'probs':
                continue

            linears[k] = nn.Linear(in_features, out_features)
            transforms[k] = torch.distributions.transform_to(v)

        self.distribution = distribution

        self.linears = nn.ModuleDict(linears)
        self.transforms = transforms

        self._outputs = defaultdict(list)

    def forward(self, x):
        kwargs = {
            k: self.transforms[k](layer(x))
            for k, layer in self.linears.items()
        }

        m = self.distribution(**kwargs)
        y = m.sample()

        for k, v in kwargs.items():
            self._outputs[f'{self.tag}.{k}'].append(v)

        self._outputs[self.tag].append(y)

        return y

    def get_outputs(self):
        outputs = {}
        for k, v in self._outputs.items():
            outputs[k] = torch.cat(v, dim=1)

        return outputs

    def reset(self):
        self._outputs = defaultdict(list)

    def calculate_loss(
        self,
        outputs,
        batch
    ) -> torch.Tensor:
        kwargs = {
            k: outputs[f'{self.tag}.{k}']
            for k in self.linears.keys()
        }

        m = self.distribution(**kwargs)

        return -torch.mean(m.log_prob(batch[self.label_tag]))


class ForecastingModule(pl.LightningModule):
    _SPECIAL_ATTRIBUTES = (
        'encoding_length',
        'decoding_length',
        'head',
        'heads',
    )

    def __init__(self):
        """Base class of all forecasting modules.
        """
        super().__init__()

        self._encoding_length = None
        self._decoding_length = None

        self._heads = None

    def __setattr__(self, name, value):
        if name in ForecastingModule._SPECIAL_ATTRIBUTES:
            return object.__setattr__(self, name, value)
        else:
            return super().__setattr__(name, value)

    @property
    def encoding_length(self) -> int:
        """Encoding length."""
        if self._encoding_length is None:
            raise NotImplementedError(
                f'Define {self.__class__.__name__}.encoding_length'
            )
        else:
            return self._encoding_length

    @encoding_length.setter
    def encoding_length(self, value: int):
        if not isinstance(value, int):
            raise TypeError(
                f'Invalid type for "encoding_length": {type(value)}'
            )
        elif value <= 0:
            raise ValueError(
                'Encoding length <= 0.'
            )
        self._encoding_length = value

    @property
    def decoding_length(self) -> int:
        if self._decoding_length is None:
            raise NotImplementedError(
                f'Define {self.__class__.__name__}.decoding_length'
            )
        else:
            return self._decoding_length

    @decoding_length.setter
    def decoding_length(self, value: int):
        if not isinstance(value, int):
            raise TypeError(
                f'Invalid type for "decoding_length": {type(value)}'
            )
        elif value <= 0:
            raise ValueError(
                'Decoding length <= 0.'
            )
        self._decoding_length = value

    @property
    def heads(self) -> list[BaseHead]:
        if self._heads is None:
            raise NotImplementedError(
                f'Define {self.__class__.__name__}.heads'
            )
        else:
            return self._heads

    @heads.setter
    def heads(self, heads: list[BaseHead]):
        if not isinstance(heads, list):
            raise TypeError(
                f'Invalid type for "heads". {type(heads)}'
            )
        elif not all(isinstance(head, BaseHead) for head in heads):
            raise TypeError(
                f'Invalid type for "heads". {[type(v) for v in heads]}'
            )

        self._heads = nn.ModuleList(heads)

    @property
    def head(self) -> BaseHead:
        if self._heads is None:
            raise NotImplementedError(
                f'Define {self.__class__.__name__}.heads'
            )
        elif len(self._heads) != 1:
            raise Exception('Multi-head model cannot use head.')
        else:
            return self._heads[0]

    @head.setter
    def head(self, head: BaseHead):
        if not isinstance(head, BaseHead):
            raise TypeError(
                f'Invalid type for "heads". {type(head)}'
            )

        self.heads = [head]

    def encode(self, inputs: dict[str, Any]) -> dict[str, Any]:
        raise NotImplementedError(
            f'Define {self.__class__.__name__}.encode()'
        )

    def decode_train(self, inputs: dict[str, Any]) -> dict[str, Any]:
        return self.decode_eval(inputs)

    def decode_eval(self, inputs: dict[str, Any]) -> dict[str, Any]:
        NotImplementedError(
            f'Define {self.__class__.__name__}.decode()'
        )

    def decode(self, inputs):
        if self.training:
            return self.decode_train(inputs)
        else:
            return self.decode_eval(inputs)

    def forward(self, inputs: dict[str, Any]) -> dict[str, Any]:
        encoder_outputs = self.encode(inputs)
        decoder_inputs = merge_dicts(
            [inputs, encoder_outputs]
        )
        outputs = self.decode(decoder_inputs)

        return outputs

    def make_chunk_specs(self):
        pass

    def calculate_loss(
        self,
        outputs: dict[str, Any],
        batch: dict[str, Any]
    ) -> dict[str, Any]:
        loss = 0
        for head in self.heads:
            loss += head.loss_weight * head.calculate_loss(outputs, batch)

        return loss

    def forward_metrics(
        self,
        outputs: dict[str, Any],
        batch: dict[str, Any],
        stage: str,
    ) -> dict[str, Any]:
        metrics = {}
        for head in self.heads:
            if not head.has_metrics:
                continue
            # It's a dictionary's update method.
            # Don't confuse with update of TorchMetric.
            metrics.update(
                # __call__ of TorchMetric.
                head.metrics(outputs=outputs, batch=batch, stage=stage)
            )

        return metrics

    def update_metrics(
        self,
        outputs: dict[str, Any],
        batch: dict[str, Any],
        stage: str,
    ) -> None:
        for head in self.heads:
            if not head.has_metrics:
                continue
            head.metrics.update(outputs=outputs, batch=batch, stage=stage)

    def compute_metrics(self, stage: str) -> dict[str, Any]:
        metrics = {}
        for head in self.heads:
            if not head.has_metrics:
                continue
            metrics.update(head.metrics.compute(stage=stage))

        return metrics

    def reset_metrics(self, stage: str) -> None:
        for head in self.heads:
            if not head.has_metrics:
                continue
            head.metrics.reset(stage=stage)

    def training_step(
        self,
        batch: dict[str, Any], batch_idx: int
    ) -> dict[str, Any]:
        outputs = self(batch)
        loss = self.calculate_loss(outputs, batch)

        # Update and evaluate metric.
        metrics = self.forward_metrics(
            outputs, batch, stage='train'
        )

        self.log('train/loss', loss)
        # Log instant metrics.
        self.log_dict(metrics)

        return loss

    def on_train_epoch_end(self) -> None:
        # Don't log epoch averaged metrics. Just reset the states.
        self.reset_metrics(stage='train')

    def validation_step(
        self,
        batch: dict[str, Any],
        batch_idx: int,
        dataloader_idx: int = 0,
    ):
        outputs = self(batch)
        loss = self.calculate_loss(outputs, batch)

        # Don't log metrics yet.
        self.update_metrics(
            outputs, batch, stage='val'
        )

        # loss will be epoch averaged.
        self.log('val/loss', loss)

    def on_validation_epoch_end(self) -> None:
        metrics = self.compute_metrics(stage='val')
        self.log_dict(metrics)
        self.reset_metrics(stage='val')

    def test_step(
        self,
        batch: dict[str, Any],
        batch_idx: int,
        dataloader_idx: int = 0,
    ):
        outputs = self(batch)
        loss = self.calculate_loss(outputs, batch)

        # Don't log metrics yet.
        self.update_metrics(
            outputs, batch, stage='test'
        )

        # loss will be epoch averaged.
        self.log('test/loss', loss)

    def on_test_epoch_end(self) -> None:
        metrics = self.compute_metrics(stage='test')
        self.log_dict(metrics)
        self.reset_metrics(stage='test')


import math
import torch
import torch.nn as nn


class Unsqueesze(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        return x.unsqueeze(self.dim)


class Permute(nn.Module):
    def __init__(self, *dims):
        super().__init__()
        self.dims = dims

    def forward(self, x):
        return x.permute(*self.dims)


class LeftPadding1D(nn.Module):
    def __init__(self, padding_size):
        super().__init__()
        self.padding_size = padding_size

    def forward(self, x):
        # x: (B, L, C).
        B = x.size(0)
        C = x.size(2)

        padding = torch.zeros(
            B, self.padding_size, C,
            dtype=x.dtype, device=x.device
        )

        y = torch.cat([padding, x], axis=1)

        return y


# Modified from:
# https://pytorch.org/tutorials/beginner/transformer_tutorial.html.
# Modifications
# 1. Dropout functionality is removed.
# 2. Changed to batch_first format.
class PositionalEncoding(nn.Module):
    def __init__(self, d_model, max_len):
        super().__init__()

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) \
            * (-math.log(10000.0) / d_model))

        pe = torch.zeros(max_len, 1, d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)

        # Change to batch_first format.
        pe = pe.permute((1, 0, 2))

        self.register_buffer('pe', pe)

    def forward(self, x):
        # x: (B, L, F).
        x = x + self.pe[:, :x.size(1), :]
        return x


import math
import numpy as np
import torch
import torch.nn as nn

from ..core import (
    ForecastingModule,
    Head,
)

from ..chunk import (
    EncodingChunkSpec,
    DecodingChunkSpec,
    LabelChunkSpec,
)

from ..layer import LeftPadding1D, Permute


class DilatedCNN(ForecastingModule):
    def __init__(
        self,
        hidden_size,
        encoding_length,
        decoding_length,
        target_names,
        nontarget_names,
        dilation_base,
        kernel_size,
        activation=nn.ELU,
        dropout_rate=0.0,
        lr=1e-3,
        optimizer: torch.optim.Optimizer = torch.optim.Adam,
        optimizer_options=None,
        loss_fn=None,
        metrics=None,
        head=None,
    ):
        super().__init__()
        self.save_hyperparameters()

        assert kernel_size >= dilation_base

        self.encoding_length = encoding_length
        self.decoding_length = decoding_length

        if optimizer_options is None:
            self.hparams.optimizer_options = {}

        if loss_fn is None:
            loss_fn = nn.MSELoss()

        n_outputs = len(target_names)
        n_features = len(nontarget_names) + n_outputs

        self.use_nontargets = n_outputs != n_features

        # Calculate the number of layers.
        # Formula is obtained from Darts.
        v = (encoding_length-1) * (dilation_base-1) / (kernel_size-1) + 1
        n_layers = math.ceil(math.log(v) / math.log(dilation_base))

        layers = []
        for i in range(n_layers):
            dilation = dilation_base**i
            padding_size = dilation * (kernel_size - 1)

            # (B, L', F)
            layers.append(LeftPadding1D(padding_size))
            # (B, F, L')
            layers.append(Permute(0, 2, 1))

            if i == 0:
                layers.append(nn.Conv1d(
                    n_features, hidden_size,
                    kernel_size=kernel_size, dilation=dilation,
                ))
            else:
                layers.append(nn.Conv1d(
                    hidden_size, hidden_size,
                    kernel_size=kernel_size, dilation=dilation,
                ))

            layers.append(activation())

            if dropout_rate > 1e-6:
                layers.append(nn.Dropout(p=dropout_rate))

            layers.append(Permute(0, 2, 1))

        # (B, H, L).
        self.body = nn.Sequential(*layers)

        if head is not None:
            self.head = head
        else:
            self.head = Head(
                tag='targets',
                output_module=nn.Linear(hidden_size, n_outputs),
                loss_fn=loss_fn,
                metrics=metrics,
            )

    def encode(self, inputs):
        # (B, L, F).
        if self.use_nontargets:
            x = torch.cat([
                inputs['encoding.targets'],
                inputs['encoding.nontargets']
            ], dim=2)
        else:
            x = inputs['encoding.targets']

        return {'x': x}

    def decode(self, inputs):
        # F = T + C.
        # (B, L, F)
        x = inputs['x']

        # (B, L, C).
        if self.use_nontargets:
            c = inputs['decoding.nontargets']

        self.head.reset()
        for i in range(self.decoding_length):
            # (B, 1, H).
            h = self.body(x)[:, -1:, :]
            # (B, 1, n_outputs).
            y = self.head(h)

            if i+1 == self.decoding_length:
                break
            # (B, 1, F).
            if self.use_nontargets:
                z = torch.cat([y, c[:, i:i+1, :]], dim=2)
            else:
                z = y

            # Concat on time dimension.
            # Remove first time step of previous input x
            # and append predicted value.
            x = torch.cat([
                x[:, 1:, :], z
            ], dim=1)

        outputs = self.head.get_outputs()

        return outputs

    def make_chunk_specs(self):
        E = self.encoding_length
        D = self.decoding_length

        chunk_specs = [
            EncodingChunkSpec(
                tag='targets',
                names=self.hparams.target_names,
                range_=(0, E),
                dtype=np.float32
            ),
            LabelChunkSpec(
                tag='targets',
                names=self.hparams.target_names,
                range_=(E, E+D),
                dtype=np.float32,
            ),
        ]

        if self.use_nontargets:
            chunk_specs += [
                EncodingChunkSpec(
                    tag='nontargets',
                    names=self.hparams.nontarget_names,
                    range_=(1, E+1),
                    dtype=np.float32,
                ),
                DecodingChunkSpec(
                    tag='nontargets',
                    names=self.hparams.nontarget_names,
                    range_=(E+1, E+D),
                    dtype=np.float32,
                ),
            ]

        return chunk_specs

    def configure_optimizers(self):
        return self.hparams.optimizer(
            self.parameters(),
            lr=self.hparams.lr,
            **self.hparams.optimizer_options,
        )


import numpy as np
import pandas as pd

from torch.utils.data import Dataset

from .chunk import BaseChunkSpec, ChunkExtractor
from .plotting import plot_chunks


class TimeSeriesDataset(Dataset):
    def __init__(self,
        data_frames: pd.DataFrame | list[pd.DataFrame],
        chunk_specs: list[BaseChunkSpec],
        return_time_index: bool = True,
    ):
        if isinstance(data_frames, pd.DataFrame):
            data_frames = [data_frames]

        self.data_frames = data_frames
        self.chunk_specs = chunk_specs
        self.return_time_index = return_time_index

        self._preprocess()

    def _preprocess(self):
        self.chunk_extractors = [
            ChunkExtractor(df, self.chunk_specs) for df in self.data_frames
        ]

        self.lengths = [
            len(df) - self.chunk_extractors[0].chunk_length + 1
            for df in self.data_frames
        ]

        self.min_start_time_index = \
            max(0, -self.chunk_extractors[0].chunk_min_t)

    def __len__(self):
        return sum(self.lengths)

    def __getitem__(self, i):
        cumsum = np.cumsum([0] + self.lengths)
        df_index = np.argmax(cumsum - i > 0) - 1

        chunk_extractor = self.chunk_extractors[df_index]
        start_time_index = i - cumsum[df_index] + self.min_start_time_index

        chunk_dict = chunk_extractor.extract(
            start_time_index, self.return_time_index
        )

        return chunk_dict

    def plot_chunks(self):
        plot_chunks(self.chunk_specs)


import copy
import numpy as np
import pandas as pd


def _merge_data_frames(
    data_frames: pd.DataFrame | list[pd.DataFrame]
) -> pd.DataFrame:
    if isinstance(data_frames, pd.DataFrame):
        data_frames = [data_frames]

    return pd.concat(data_frames).reset_index(drop=True)


class ColumnTransformer:
    def __init__(
        self,
        transformer_dict=None,
        transformer_tuples=None,
    ):
        if transformer_dict is None and transformer_tuples is None:
            raise Exception(
                'One of transformer_dict or transformer_tuples has to be given.'
            )

        if (transformer_dict is not None) and (transformer_tuples is not None):
            raise Exception(
                'Both transformer_dict or transformer_tuples are set.'
            )

        if transformer_dict is not None:
            self.transformer_dict = transformer_dict

        if transformer_tuples is not None:
            self.transformer_dict = {}
            for transformer, names in transformer_tuples:
                for name in names:
                    self.transformer_dict[name] = copy.deepcopy(transformer)

    def _apply_to_single_feature(self, series, func):
        values = series.values.reshape(-1, 1)
        return_value = func(values)
        if isinstance(return_value, np.ndarray):
            return return_value.reshape(-1)
        else:
            return return_value

    def _get_valid_names(self, names):
        valid_name_set = set(self.transformer_dict.keys()) & set(names)
        # Do not use list(valid_name_set) to preserve order of elements.
        return [name for name in names if name in valid_name_set]

    def fit(
        self,
        data_frames: pd.DataFrame | list[pd.DataFrame]
    ) -> None:
        df = _merge_data_frames(data_frames)

        for name in self._get_valid_names(df.columns):
            transformer = self.transformer_dict[name]

            self._apply_to_single_feature(
                df[name], transformer.fit
            )

    def transform(
        self,
        data_frames: pd.DataFrame | list[pd.DataFrame]
    ) -> pd.DataFrame | list[pd.DataFrame]:
        single_df = isinstance(data_frames, pd.DataFrame)

        if single_df:
            data_frames = [data_frames]

        dfs = []
        for df in data_frames:
            data = {}
            for name in self._get_valid_names(df.columns):
                transformer = self.transformer_dict[name]

                data[name] = self._apply_to_single_feature(
                    df[name], transformer.transform
                )

            dfs.append(pd.DataFrame(data=data, index=df.index))

        if single_df:
            return dfs[0]
        else:
            return dfs

    def fit_transform(
        self, data_frames: pd.DataFrame | list[pd.DataFrame]
    ) -> pd.DataFrame | list[pd.DataFrame]:
        self.fit(data_frames)
        return self.transform(data_frames)

    def inverse_transform(
        self, data_frames: pd.DataFrame | list[pd.DataFrame]
    ) -> pd.DataFrame | list[pd.DataFrame]:
        single_df = isinstance(data_frames, pd.DataFrame)

        if single_df:
            data_frames = [data_frames]

        dfs = []
        for df in data_frames:
            data = {}
            for name in self._get_valid_names(df.columns):
                transformer = self.transformer_dict[name]

                data[name] = self._apply_to_single_feature(
                    df[name], transformer.inverse_transform
                )

            dfs.append(pd.DataFrame(data=data, index=df.index))

        if single_df:
            return dfs[0]
        else:
            return dfs


from .core import (
    ForecastingModule,
    MetricModule,
    BaseHead,
    Head,
    DistributionHead,
)

from .chunk import (
    BaseChunkSpec,
    EncodingChunkSpec,
    DecodingChunkSpec,
    LabelChunkSpec,
    ChunkExtractor,
)

from .dataset import (
    TimeSeriesDataset,
)

from .transform import (
    ColumnTransformer,
)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deep_time_series', 'deep_time_series.model']

package_data = \
{'': ['*']}

install_requires = \
['lightning>=2.0.0,<3.0.0',
 'matplotlib>=3.7.1,<4.0.0',
 'numpy>=1.24.2,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'rich>=13.3.2,<14.0.0',
 'scikit-learn>=1.2.2,<2.0.0',
 'torch>=2.0.0,<3.0.0',
 'torchaudio>=2.0.1,<3.0.0',
 'torchvision>=0.15.1,<0.16.0',
 'xarray>=2023.2.0,<2024.0.0']

setup_kwargs = {
    'name': 'deep-time-series',
    'version': '0.1.1',
    'description': 'Deep learning library for time series forecasting based on PyTorch.',
    'long_description': '# DeepTimeSeries\n\nLast update Oct.19, 2022\n\nDeep learning library for time series forecasting based on PyTorch.\nIt is under development and the first released version will be announced soon.\n\n## Why DeepTimeSeries?\n\nDeepTimeSeries is inspired by libraries such as ``Darts`` and\n``PyTorch Forecasting``. So why was DeepTimeSeries developed?\n\nThe design philosophy of DeepTimeSeries is as follows:\n\n**We present logical guidelines for designing various deep learning models for\ntime series forecasting**\n\nOur main target users are intermediate-level users who need to develop\ndeep learning models for time series prediction.\nWe provide solutions to many problems that deep learning modelers face\nbecause of the uniqueness of time series data.\n\nWe additionally implement a high-level API, which allows comparatively beginners\nto use models that have already been implemented.\n\n## Supported Models\n\n| Model       | Target features | Non-target features | Deterministic | Probabilistic |\n| ----------- | --------------- | ------------------- | ------------- | ------------- |\n| MLP         | o               | o                   | o             | o             |\n| Dilated CNN | o               | o                   | o             | o             |\n| Vanilla RNN | o               | o                   | o             | o             |\n| LSTM        | o               | o                   | o             | o             |\n| GRU         | o               | o                   | o             | o             |\n| Transformer | o               | o                   | o             | o             |\n\n## Documentation\n\nhttps://bet-lab.github.io/DeepTimeSeries/\n',
    'author': 'Sangwon',
    'author_email': 'lsw91.main@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import numpy as np
import torch
import torch.nn as nn

from ..core import (
    ForecastingModule,
    Head,
)

from ..chunk import (
    EncodingChunkSpec,
    DecodingChunkSpec,
    LabelChunkSpec,
)

from ..layer import (
    Unsqueesze,
)


class MLP(ForecastingModule):
    def __init__(
        self,
        hidden_size,
        encoding_length,
        decoding_length,
        target_names,
        nontarget_names,
        n_hidden_layers,
        activation=nn.ELU,
        dropout_rate=0.0,
        lr=1e-3,
        optimizer: torch.optim.Optimizer = torch.optim.Adam,
        optimizer_options=None,
        loss_fn=None,
        metrics=None,
        head=None,
    ):
        super().__init__()
        self.save_hyperparameters()

        self.encoding_length = encoding_length
        self.decoding_length = decoding_length

        if optimizer_options is None:
            self.hparams.optimizer_options = {}

        if loss_fn is None:
            loss_fn = nn.MSELoss()

        n_outputs = len(target_names)
        n_features = len(nontarget_names) + n_outputs

        self.use_nontargets = n_outputs != n_features

        layers = [
            nn.Flatten(),
            nn.Linear(n_features*encoding_length, hidden_size),
            activation()
        ]
        for i in range(n_hidden_layers):
            if dropout_rate > 1e-6:
                layers.append(nn.Dropout(p=dropout_rate))
            layers.append(nn.Linear(hidden_size, hidden_size))
            layers.append(activation())

        if dropout_rate > 1e-6:
            layers.append(nn.Dropout(p=dropout_rate))

        layers.append(Unsqueesze(1))

        self.body = nn.Sequential(*layers)

        if head is not None:
            self.head = head
        else:
            self.head = Head(
                tag='targets',
                output_module=nn.Linear(hidden_size, n_outputs),
                loss_fn=loss_fn,
                metrics=metrics,
            )

    def encode(self, inputs):
        # (B, L, F).
        if self.use_nontargets:
            x = torch.cat([
                inputs['encoding.targets'],
                inputs['encoding.nontargets']
            ], dim=2)
        else:
            x = inputs['encoding.targets']

        return {'x': x}

    def decode(self, inputs):
        # F = T + C.
        # (B, L, F)
        x = inputs['x']

        # (B, L, C).
        if self.use_nontargets:
            c = inputs['decoding.nontargets']

        B = x.size(0)
        EL = self.encoding_length

        self.head.reset()
        for i in range(self.decoding_length):
            # (B, L*F)
            # x = x.view(B, -1)
            # (B, 1, n_outputs).
            y = self.body(x)
            # y = y.unsqueeze(1)
            y = self.head(y)

            if i+1 == self.decoding_length:
                break
            # (B, 1, F).
            if self.use_nontargets:
                z = torch.cat([y, c[:, i:i+1, :]], dim=2)
            else:
                z = y
            # z = y
            # (B, EL, F).
            x = x.view(B, EL, -1)
            # (B, L, F).
            x = torch.cat([
                x[:, 1:, :], z
            ], dim=1)

        outputs = self.head.get_outputs()

        return outputs

    def make_chunk_specs(self):
        E = self.encoding_length
        D = self.decoding_length

        chunk_specs = [
            EncodingChunkSpec(
                tag='targets',
                names=self.hparams.target_names,
                range_=(0, E),
                dtype=np.float32
            ),
            LabelChunkSpec(
                tag='targets',
                names=self.hparams.target_names,
                range_=(E, E+D),
                dtype=np.float32,
            ),
        ]

        if self.use_nontargets:
            chunk_specs += [
                EncodingChunkSpec(
                    tag='nontargets',
                    names=self.hparams.nontarget_names,
                    range_=(1, E+1),
                    dtype=np.float32,
                ),
                DecodingChunkSpec(
                    tag='nontargets',
                    names=self.hparams.nontarget_names,
                    range_=(E+1, E+D),
                    dtype=np.float32,
                ),
            ]

        return chunk_specs

    def configure_optimizers(self):
        return self.hparams.optimizer(
            self.parameters(),
            lr=self.hparams.lr,
            **self.hparams.optimizer_options,
        )